from django.test import TestCase, override_settings


class TestHttp2ServerPushRedirectMiddleware(TestCase):
    # Must set DEBUG to True for the logger to trigger:
    # https://docs.djangoproject.com/en/2.2/topics/http/middleware/#marking-middleware-as-unused
    @override_settings(DEBUG=True, SECURE_PROXY_SSL_HEADER=None)
    def test_disabled_when_secure_proxy_ssl_header_is_not_configured(self):
        with self.assertLogs("django.request", level="DEBUG") as cm:
            self.client.get("/")
        self.assertIn(
            "DEBUG:django.request:MiddlewareNotUsed"
            "('push_redirect.middleware.Http2ServerPushRedirectMiddleware'):"
            " SECURE_PROXY_SSL_HEADER is not configured",
            cm.output,
        )

    def test_non_secure_request(self):
        """
        Does not add the preload header redirects requested over HTTP
        """
        response = self.client.get("/")
        self.assertRedirects(response, "/hello", fetch_redirect_response=False)
        self.assertNotIn("Link", response)

    def test_secure_request(self):
        """
        Adds the preload header to redirects requested over HTTPS
        """
        response = self.client.get("/", secure=True)
        self.assertRedirects(response, "/hello", fetch_redirect_response=False)
        self.assertIn("Link", response)
        self.assertEqual("</hello>; as=document; rel=preload", response["Link"])

    def test_post_requests(self):
        """
        Does not add the preload header to redirects following
        a POST (or any other non-GET) request
        """
        response = self.client.post("/", secure=True)
        self.assertRedirects(response, "/hello", fetch_redirect_response=False)
        self.assertNotIn("Link", response)

    def test_non_redirect_responses(self):
        """
        Does not add the preload header to a 200 (OK) response
        """
        response = self.client.get("/hello/world/", secure=True)
        self.assertEqual(200, response.status_code)
        self.assertNotIn("Link", response)

    def test_custom_redirect_responses(self):
        """
        Does not add the preload header to a custom
        (non HttpResponseRedirectBase) response
        """
        response = self.client.get("/custom-redirect/", secure=True)
        # Cannot use assertRedirects as it also requires the url attribute
        # added by HttpResponseRedirectBae
        # self.assertRedirects(response, "/custom-redirect/redirected/", fetch_redirect_response=False)
        self.assertIn("Location", response)
        self.assertEqual('/custom-redirect/redirected/', response['Location'])
        self.assertNotIn("Link", response)

    def test_existing_link_header_in_responses(self):
        """
        Does not overwrite existing Link headers
        """
        response = self.client.get("/custom-link/", secure=True)
        self.assertRedirects(
            response, "https://www.google.com/", fetch_redirect_response=False
        )
        self.assertIn("Link", response)
        self.assertEqual("<https://www.google.com/>; rel=preconnect", response["Link"])

    def test_external_redirect_response(self):
        """
        Does not add the preload header to a external redirect
        """
        response = self.client.get("/google/", secure=True)
        self.assertRedirects(
            response, "https://www.google.com/", fetch_redirect_response=False
        )
        self.assertNotIn("Link", response)

    def test_common_middleware(self):
        """
        Adds the preload header to redirects created by Django's
        CommonMiddleware (APPEND_SLASH)
        """
        response = self.client.get("/hello/moon", secure=True)
        self.assertRedirects(response, "/hello/moon/", 301, fetch_redirect_response=False)
        self.assertIn("Link", response)
        self.assertEqual("</hello/moon/>; as=document; rel=preload", response["Link"])



import os

SECRET_KEY = "not-that-secret"

DEBUG = True

ALLOWED_HOSTS = ["*"]

INSTALLED_APPS = []

MIDDLEWARE = [
    # Must come first to rewrite the response of CommonMiddleware
    "push_redirect.middleware.Http2ServerPushRedirectMiddleware",
    "django.middleware.common.CommonMiddleware",
]

ROOT_URLCONF = "tests.urls"

TEMPLATES = [{"BACKEND": "django.template.backends.django.DjangoTemplates"}]

WSGI_APPLICATION = "tests.wsgi.application"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(os.path.dirname(__file__), "db.sqlite3"),
    }
}

# Trust proxy headers

SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTO", "https")



from django.http import HttpResponse, HttpResponseRedirect
from django.urls import path, reverse
from django.utils import lorem_ipsum


class CustomRedirectResponse(HttpResponse):
    status_code = 302

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self['Location'] = '/custom-redirect/redirected/'


class CustomLinkResponse(HttpResponseRedirect):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self['Link'] = '<https://www.google.com/>; rel=preconnect'


urlpatterns = [
    path("", lambda request: HttpResponseRedirect(reverse("hello").rstrip("/"))),
    path(
        "hello/",
        lambda request: HttpResponseRedirect(
            reverse("hello", kwargs={"name": "world"}).rstrip("/")
        ),
        name="hello",
    ),
    path(
        "hello/<slug:name>/",
        lambda request, name: HttpResponse(
            f"<!doctype html>"
            f"<title>Hello {name}</title>"
            f"<h1>Hello {name}</h1>"
            f'{"<p>".join(lorem_ipsum.paragraphs(1))}'
        ),
        name="hello",
    ),
    path(
        "custom-redirect/",
        lambda request: CustomRedirectResponse(),
        name="custom-redirect",
    ),
    path(
        "custom-link/",
        lambda request: CustomLinkResponse("https://www.google.com/"),
        name="custom-redirect",
    ),
    path(
        "google/",
        lambda request: HttpResponseRedirect("https://www.google.com/"),
        name="google",
    ),
]



from django.conf import settings
from django.core.exceptions import MiddlewareNotUsed
from django.utils.http import url_has_allowed_host_and_scheme


class Http2ServerPushRedirectMiddleware:
    redirect_status_codes = {301, 302, 303, 307, 308}

    def __init__(self, get_response):
        if not settings.SECURE_PROXY_SSL_HEADER:
            # For Django to be able to detect if a request is secure
            # some configuration is required, e.g..
            #
            # The nginx proxy should set:
            #
            #   proxy_set_header X-Forwarded-Proto $scheme;
            #
            # The Django project should set:
            #
            #   SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')
            raise MiddlewareNotUsed("SECURE_PROXY_SSL_HEADER is not configured")
        self.get_response = get_response

    def should_preload(self, request, response):
        if not request.is_secure():
            # HTTP/2 requires SSL/TLS.
            return False

        if not request.method == "GET":
            # Err on the side of caution by only preloading GET requests
            return False

        if response.status_code not in self.redirect_status_codes:
            # Not a redirect
            return False

        if not hasattr(response, "url"):
            # The url attribute is added by the HttpResponseRedirectBase
            # base class. This response does not have it so ignore it.
            return False

        if response.has_header("Link"):
            # There already is a Link header, don't overwrite it
            return False

        # Preload is enabled if the redirect url is on the same host
        # as the request and using the same protocol (https)
        return url_has_allowed_host_and_scheme(
            response.url, allowed_hosts={request.get_host()}, require_https=True
        )

    def __call__(self, request):
        response = self.get_response(request)
        if self.should_preload(request, response):
            # Response can be preloaded, add Link header
            response["Link"] = f"<{response.url}>; as=document; rel=preload"
        return response
